from .._core_exceptions import LograderError


class LograderStudentBuildError(LograderError):
    """
    This is the base exception class for all exceptions raised
    by the `lograder.dispatch` module, for easy error handling.
    """

    pass



import sys
from datetime import datetime
from pathlib import Path
from typing import List, Optional

from ...common.utils import random_name
from ...static import LograderBasicConfig
from ..common import (
    CLIBuilder,
    DispatcherInterface,
    ExecutableBuildResults,
    ExecutableRunner,
    PreprocessorInterface,
    PreprocessorResults,
    RuntimeResults,
    TrivialPreprocessor,
)
from ..common.file_operations import bfs_walk, is_cxx_source_file
from ..common.types import AssignmentMetadata


class CxxSourceDispatcher(CLIBuilder, ExecutableRunner, DispatcherInterface):

    def __init__(
        self,
        *,
        assignment_name: str,
        assignment_authors: List[str],
        assignment_description: str,
        assignment_due_date: datetime,
        project_root: Path = LograderBasicConfig.DEFAULT_SUBMISSION_PATH,
        preprocessor: PreprocessorInterface = TrivialPreprocessor(),
    ):
        super().__init__(build_type="cmake")
        self._metadata = AssignmentMetadata(
            assignment_name=assignment_name,
            assignment_authors=assignment_authors,
            assignment_description=assignment_description,
            assignment_due_date=assignment_due_date,
        )

        self._project_root: Path = Path(project_root)
        self._build_directory: Path = self._project_root / "build"
        if not self._build_directory.exists():
            self._build_directory = self._project_root

        self._executable_path: Optional[Path] = None
        self._preprocessor = preprocessor

    def build(self) -> ExecutableBuildResults:
        source_files: List[Path] = []
        for file in bfs_walk(self._project_root):
            if is_cxx_source_file(file):
                source_files.append(file)

        cmd: List[str | Path] = [
            "g++",
            *LograderBasicConfig.DEFAULT_CXX_COMPILATION_FLAGS,
            f"-std={LograderBasicConfig.DEFAULT_CXX_STANDARD}",
            "-o",
            self.get_executable_path(),
            *source_files,
        ]
        output = self.run_cmd(cmd)
        if self.is_build_error():
            return self.get_build_error_output()

        return ExecutableBuildResults(
            executable=self.get_executable_path(), output=output
        )

    def get_build_directory(self) -> Path:
        return self._build_directory

    def get_project_root(self) -> Path:
        return self._project_root

    def get_executable_path(self) -> Path:
        if self._executable_path is not None:
            return self._executable_path
        while True:
            executable_name = (
                self.get_build_directory() / (random_name() + ".exe")
                if sys.platform.startswith("win")
                else self.get_build_directory() / random_name()
            )
            if not executable_name.exists():
                self._executable_path = executable_name
                return self._executable_path

    def metadata(self) -> AssignmentMetadata:
        return self._metadata

    def preprocess(self) -> PreprocessorResults:
        return self._preprocessor.preprocess()

    def run_tests(self) -> RuntimeResults:
        return self.run_tests_auto()

    def get_executable(self) -> List[str | Path]:
        return [self.get_executable_path()]



from typing import Any, Dict, List, Optional

from pydantic import BaseModel, Field, computed_field, model_validator

from ..common import TextFormat, Visibility
from .leaderboard import LeaderboardJSON
from .test_case import TestCaseJSON


class AssignmentJSON(BaseModel):
    score: Optional[float] = None
    output: Optional[str] = None  # Text relevant to the entire submission
    output_format: Optional[TextFormat] = "ansi"
    test_output_format: Optional[TextFormat] = "ansi"
    visibility: Optional[Visibility] = "visible"
    extra_data: Optional[Dict[str, Any]] = Field(default_factory=dict)
    tests: Optional[List[TestCaseJSON]]
    leaderboard: Optional[List[LeaderboardJSON]] = None

    @model_validator(mode="after")
    def check_score_existence(self):
        if not self.tests:
            if self.score is None:
                raise ValueError(
                    "If there are no tests specified, please pass `score` to Grade object."
                )
        elif all([test.is_scored for test in self.tests]):
            if (
                self.score is not None
            ):  # Technically, you are allowed to overwrite, but that's stupid.
                raise ValueError(
                    "You have specified tests with `tests`, but you are overwriting the score of the Grade object."
                )
        return self

    @property
    @computed_field
    def execution_time(self) -> float:
        total = 0.0
        if self.tests is not None:
            for test in self.tests:
                if test.execution_time is not None:
                    total += test.execution_time
        return total



from typing import Callable, Generator, List

from ..registry import TestRegistry
from ..test import ExecutableOutputComparisonTest
from .types import (
    FlaggedTestCaseProtocol,
    FlaggedWeightedTestCaseProtocol,
    TestCase,
    TestCaseProtocol,
    WeightedTestCaseProtocol,
)


def make_tests_from_generator(
    generator: Callable[[], Generator[TestCase, None, None]],
) -> None:
    generated_tests: List[ExecutableOutputComparisonTest] = []
    for test_case in generator():
        if isinstance(test_case, WeightedTestCaseProtocol):
            test = ExecutableOutputComparisonTest(
                name=test_case.get_name(),
                input=test_case.get_input(),
                expected_output=test_case.get_expected_output(),
                weight=test_case.get_weight(),
                flags=[],
            )
            generated_tests.append(test)
        elif isinstance(test_case, TestCaseProtocol):
            test = ExecutableOutputComparisonTest(
                name=test_case.get_name(),
                input=test_case.get_input(),
                expected_output=test_case.get_expected_output(),
                weight=1.0,
                flags=[],
            )
            generated_tests.append(test)
        elif isinstance(test_case, FlaggedWeightedTestCaseProtocol):
            test = ExecutableOutputComparisonTest(
                name=test_case.get_name(),
                input=test_case.get_input(),
                expected_output=test_case.get_expected_output(),
                weight=test_case.get_weight(),
                flags=test_case.get_flags(),
            )
            generated_tests.append(test)
        elif isinstance(test_case, FlaggedTestCaseProtocol):
            test = ExecutableOutputComparisonTest(
                name=test_case.get_name(),
                input=test_case.get_input(),
                expected_output=test_case.get_expected_output(),
                weight=1.0,
                flags=test_case.get_flags(),
            )
            generated_tests.append(test)
        elif isinstance(test_case, dict):
            if "weight" in test_case:
                weight = test_case["weight"]
            else:
                weight = 1.0
            if "flags" in test_case:
                flags = test_case["flags"]
            else:
                flags = []

            test = ExecutableOutputComparisonTest(
                name=test_case["name"],
                input=test_case["input"],
                expected_output=test_case["expected_output"],
                weight=weight,
                flags=flags,
            )
            generated_tests.append(test)
        else:
            raise ValueError(
                f"`generator` passed to `make_tests_from_generator` produced a type, `{type(test_case)}`, which is not supported."
            )

    TestRegistry.extend(generated_tests)



from datetime import datetime
from pathlib import Path
from typing import List

from ...static import LograderBasicConfig
from ..common.file_operations import bfs_walk, is_cmake_file, is_makefile_file
from ..common.interface import (
    DispatcherInterface,
    ExecutableBuildResults,
    PreprocessorInterface,
    PreprocessorResults,
    RuntimeResults,
)
from ..common.templates import TrivialPreprocessor
from ..common.types import AssignmentMetadata, ProjectType
from ..cpp import CMakeDispatcher, CxxSourceDispatcher
from .makefile import MakefileDispatcher


def detect_project_type(project_root: Path) -> ProjectType:
    for file in bfs_walk(project_root):
        if is_cmake_file(file):
            return "cmake"
        if is_makefile_file(file):
            return "makefile"
    return "cxx-source"


class ProjectDispatcher(DispatcherInterface):
    def __init__(
        self,
        *,
        assignment_name: str,
        assignment_authors: List[str],
        assignment_description: str,
        assignment_due_date: datetime,
        project_root: Path = LograderBasicConfig.DEFAULT_SUBMISSION_PATH,
        preprocessor: PreprocessorInterface = TrivialPreprocessor(),
    ):
        super().__init__()
        self._metadata = AssignmentMetadata(
            assignment_name=assignment_name,
            assignment_authors=assignment_authors,
            assignment_description=assignment_description,
            assignment_due_date=assignment_due_date,
        )

        project_type: ProjectType = detect_project_type(project_root)
        self._internal_project: DispatcherInterface

        if project_type == "cmake":
            self._internal_project = CMakeDispatcher(
                assignment_name=assignment_name,
                assignment_authors=assignment_authors,
                assignment_description=assignment_description,
                assignment_due_date=assignment_due_date,
                project_root=project_root,
                preprocessor=preprocessor,
            )
        elif project_type == "makefile":
            self._internal_project = MakefileDispatcher(
                assignment_name=assignment_name,
                assignment_authors=assignment_authors,
                assignment_description=assignment_description,
                assignment_due_date=assignment_due_date,
                project_root=project_root,
                preprocessor=preprocessor,
            )
        else:
            self._internal_project = CxxSourceDispatcher(
                assignment_name=assignment_name,
                assignment_authors=assignment_authors,
                assignment_description=assignment_description,
                assignment_due_date=assignment_due_date,
                project_root=project_root,
                preprocessor=preprocessor,
            )

    def metadata(self) -> AssignmentMetadata:
        return self._internal_project.metadata()

    def preprocess(self) -> PreprocessorResults:
        return self._internal_project.preprocess()

    def build(self) -> ExecutableBuildResults:
        return self._internal_project.build()

    def run_tests(self) -> RuntimeResults:
        return self._internal_project.run_tests()



from .test_maker import TestCaseTemplate, make_tests_from_template
from .types import TemplateSubstitution, TSub

__all__ = [
    "TemplateSubstitution",
    "TSub",
    "TestCaseTemplate",
    "make_tests_from_template",
]



import shutil
import sys
from datetime import datetime
from pathlib import Path

import pytest

from lograder.dispatch import AssignmentSummary, ProjectDispatcher
from lograder.output.formatters.default import (
    DefaultExecutableTestCaseFormatter,
)
from lograder.tests import make_tests_from_files


@pytest.mark.description('Testing C++ Source Build of correct "Hello World!" project.')
def test_project_1(tmp_path):
    here = Path(__file__).parent
    src_project_dir = here / "test-projects" / "project-1"
    shutil.copytree(src_project_dir, tmp_path, dirs_exist_ok=True)

    assert (tmp_path / "hello_world.cpp").is_file()
    assert (tmp_path / "expected_output.txt").is_file()

    make_tests_from_files(
        names=['Testing C++ Source Build of correct "Hello World!" project.'],
        expected_output_files=[tmp_path / "expected_output.txt"],
        input_strs=[""],
    )

    assignment = ProjectDispatcher(
        assignment_name="Lograder Unit Testing",
        assignment_authors=["Logan Dapp"],
        assignment_description="for `pytest`-ing of the Gradescope Autograder library.",
        assignment_due_date=datetime(year=2030, month=12, day=12),
        project_root=tmp_path,
    )
    metadata = assignment.metadata()
    preprocessor_results = assignment.preprocess()
    builder_results = assignment.build()
    runtime_results = assignment.run_tests()
    summary = AssignmentSummary(
        metadata=metadata,
        preprocessor_output=preprocessor_results.get_output(),
        build_output=builder_results.get_output(),
        test_cases=runtime_results.get_test_cases(),
    )
    print(summary.get_assignment_text())
    for test in runtime_results.get_test_cases():
        print(DefaultExecutableTestCaseFormatter().format(test))
        assert test.get_successful()


@pytest.mark.description('Testing CMake Build of correct "Hello World!" project.')
def test_project_2(tmp_path):
    here = Path(__file__).parent
    src_project_dir = here / "test-projects" / "project-2"
    shutil.copytree(src_project_dir, tmp_path, dirs_exist_ok=True)

    assert (tmp_path / "main.cpp").is_file()
    assert (tmp_path / "CMakeLists.txt").is_file()
    assert (tmp_path / "expected_output.txt").is_file()

    make_tests_from_files(
        names=['Testing CMake Build of correct "Hello World!" project.'],
        expected_output_files=[tmp_path / "expected_output.txt"],
        input_strs=[""],
    )

    assignment = ProjectDispatcher(
        assignment_name="Lograder Unit Testing",
        assignment_authors=["Logan Dapp"],
        assignment_description="for `pytest`-ing of the Gradescope Autograder library.",
        assignment_due_date=datetime(year=2030, month=12, day=12),
        project_root=tmp_path,
    )
    metadata = assignment.metadata()
    preprocessor_results = assignment.preprocess()
    builder_results = assignment.build()
    runtime_results = assignment.run_tests()
    summary = AssignmentSummary(
        metadata=metadata,
        preprocessor_output=preprocessor_results.get_output(),
        build_output=builder_results.get_output(),
        test_cases=runtime_results.get_test_cases(),
    )
    print(summary.get_assignment_text())
    for test in runtime_results.get_test_cases():
        print(DefaultExecutableTestCaseFormatter().format(test))
        assert test.get_successful()


if not sys.platform.startswith("win"):

    @pytest.mark.description(
        'Testing Makefile Build of correct "Hello World!" project.'
    )
    def test_project_3(tmp_path):
        here = Path(__file__).parent
        src_project_dir = here / "test-projects" / "project-3"
        shutil.copytree(src_project_dir, tmp_path, dirs_exist_ok=True)

        assert (tmp_path / "main.cpp").is_file()
        assert (tmp_path / "Makefile").is_file()
        assert (tmp_path / "expected_output.txt").is_file()

        make_tests_from_files(
            names=['Testing Makefile Build of correct "Hello World!" project.'],
            expected_output_files=[tmp_path / "expected_output.txt"],
            input_strs=[""],
        )

        assignment = ProjectDispatcher(
            assignment_name="Lograder Unit Testing",
            assignment_authors=["Logan Dapp"],
            assignment_description="for `pytest`-ing of the Gradescope Autograder library.",
            assignment_due_date=datetime(year=2030, month=12, day=12),
            project_root=tmp_path,
        )
        metadata = assignment.metadata()
        preprocessor_results = assignment.preprocess()
        builder_results = assignment.build()
        runtime_results = assignment.run_tests()
        summary = AssignmentSummary(
            metadata=metadata,
            preprocessor_output=preprocessor_results.get_output(),
            build_output=builder_results.get_output(),
            test_cases=runtime_results.get_test_cases(),
        )
        print(summary.get_assignment_text())
        for test in runtime_results.get_test_cases():
            print(DefaultExecutableTestCaseFormatter().format(test))
            assert test.get_successful()


@pytest.mark.description('Testing C++ Source Build of bad "Hello World!" project.')
def test_project_4(tmp_path):
    here = Path(__file__).parent
    src_project_dir = here / "test-projects" / "project-4"
    shutil.copytree(src_project_dir, tmp_path, dirs_exist_ok=True)

    assert (tmp_path / "bad_hello_world.cpp").is_file()
    assert (tmp_path / "expected_output.txt").is_file()

    make_tests_from_files(
        names=['Testing C++ Source Build of bad "Hello World!" project.'],
        expected_output_files=[tmp_path / "expected_output.txt"],
        input_strs=[""],
    )

    assignment = ProjectDispatcher(
        assignment_name="Lograder Unit Testing",
        assignment_authors=["Logan Dapp"],
        assignment_description="for `pytest`-ing of the Gradescope Autograder library.",
        assignment_due_date=datetime(year=2030, month=12, day=12),
        project_root=tmp_path,
    )
    metadata = assignment.metadata()
    preprocessor_results = assignment.preprocess()
    builder_results = assignment.build()
    runtime_results = assignment.run_tests()
    summary = AssignmentSummary(
        metadata=metadata,
        preprocessor_output=preprocessor_results.get_output(),
        build_output=builder_results.get_output(),
        test_cases=runtime_results.get_test_cases(),
    )
    print(summary.get_assignment_text())
    for test in runtime_results.get_test_cases():
        print(DefaultExecutableTestCaseFormatter().format(test))
        assert test.get_successful() is False


@pytest.mark.description('Testing CMake Build of bad "Hello World!" project.')
def test_project_5(tmp_path):
    here = Path(__file__).parent
    src_project_dir = here / "test-projects" / "project-5"
    shutil.copytree(src_project_dir, tmp_path, dirs_exist_ok=True)

    assert (tmp_path / "bad_main.cpp").is_file()
    assert (tmp_path / "CMakeLists.txt").is_file()
    assert (tmp_path / "expected_output.txt").is_file()

    make_tests_from_files(
        names=['Testing CMake Build of bad "Hello World!" project.'],
        expected_output_files=[tmp_path / "expected_output.txt"],
        input_strs=[""],
    )

    assignment = ProjectDispatcher(
        assignment_name="Lograder Unit Testing",
        assignment_authors=["Logan Dapp"],
        assignment_description="for `pytest`-ing of the Gradescope Autograder library.",
        assignment_due_date=datetime(year=2030, month=12, day=12),
        project_root=tmp_path,
    )
    metadata = assignment.metadata()
    preprocessor_results = assignment.preprocess()
    builder_results = assignment.build()
    runtime_results = assignment.run_tests()
    summary = AssignmentSummary(
        metadata=metadata,
        preprocessor_output=preprocessor_results.get_output(),
        build_output=builder_results.get_output(),
        test_cases=runtime_results.get_test_cases(),
    )
    print(summary.get_assignment_text())
    for test in runtime_results.get_test_cases():
        print(DefaultExecutableTestCaseFormatter().format(test))
        assert test.get_successful() is False


if not sys.platform.startswith("win"):

    @pytest.mark.description('Testing Makefile Build of bad "Hello World!" project.')
    def test_project_6(tmp_path):
        here = Path(__file__).parent
        src_project_dir = here / "test-projects" / "project-6"
        shutil.copytree(src_project_dir, tmp_path, dirs_exist_ok=True)

        assert (tmp_path / "bad_main.cpp").is_file()
        assert (tmp_path / "Makefile").is_file()
        assert (tmp_path / "expected_output.txt").is_file()

        make_tests_from_files(
            names=['Testing Makefile Build of bad "Hello World!" project.'],
            expected_output_files=[tmp_path / "expected_output.txt"],
            input_strs=[""],
        )

        assignment = ProjectDispatcher(
            assignment_name="Lograder Unit Testing",
            assignment_authors=["Logan Dapp"],
            assignment_description="for `pytest`-ing of the Gradescope Autograder library.",
            assignment_due_date=datetime(year=2030, month=12, day=12),
            project_root=tmp_path,
        )
        metadata = assignment.metadata()
        preprocessor_results = assignment.preprocess()
        builder_results = assignment.build()
        runtime_results = assignment.run_tests()
        summary = AssignmentSummary(
            metadata=metadata,
            preprocessor_output=preprocessor_results.get_output(),
            build_output=builder_results.get_output(),
            test_cases=runtime_results.get_test_cases(),
        )
        print(summary.get_assignment_text())
        for test in runtime_results.get_test_cases():
            print(DefaultExecutableTestCaseFormatter().format(test))
            assert test.get_successful() is False



from abc import ABC
from pathlib import Path
from typing import List, Optional

from ....static import LograderMessageConfig
from ..file_operations import run_cmd
from ..interface import BuilderInterface, BuilderOutput, ExecutableBuildResults
from ..types import ProjectType


class CLIBuilder(BuilderInterface, ABC):
    def __init__(self, build_type: ProjectType):
        super().__init__()
        self._commands: List[List[str | Path]] = []
        self._stdout: List[str] = []
        self._stderr: List[str] = []
        self._build_type = build_type

    def get_commands(self) -> List[List[str | Path]]:
        return self._commands

    def get_build_type(self) -> ProjectType:
        return self._build_type

    def get_stdout(self) -> List[str]:
        return self._stdout

    def get_stderr(self) -> List[str]:
        return self._stderr

    def get_build_error_output(self) -> ExecutableBuildResults:
        return ExecutableBuildResults(
            executable=LograderMessageConfig.DEFAULT_BUILD_ERROR_EXECUTABLE_NAME,
            output=BuilderOutput(
                commands=self.get_commands(),
                stdout=self.get_stdout(),
                stderr=self.get_stderr(),
                project_type=self.get_build_type(),
            ),
        )

    def run_cmd(
        self, cmd: List[str | Path], working_directory: Optional[Path] = None
    ) -> BuilderOutput:
        result = run_cmd(
            cmd,
            commands=self._commands,
            stdout=self._stdout,
            stderr=self._stderr,
            working_directory=working_directory,
        )
        if result.returncode != 0:
            self.set_build_error(True)
        return BuilderOutput(
            commands=self._commands,
            stdout=self._stdout,
            stderr=self._stderr,
            project_type=self._build_type,
        )



from typing import Optional

from pydantic import BaseModel, field_validator

from ..common import AscendingOrder


class LeaderboardJSON(BaseModel):
    name: str
    value: float | str
    order: Optional[AscendingOrder] = None

    @field_validator("value")
    def validate_value(cls, v):
        if isinstance(v, str) and any([char != "*" for char in v]):
            raise ValueError(
                "If passing a string for value, it must be made entirely of the character, '*'."
            )
        return v



import subprocess
import sys
from collections import deque
from pathlib import Path
from typing import List, Optional

from ...static.basicconfig import LograderBasicConfig


def bfs_walk(root: Path):  # pathlib defaults to dfs; must implement bfs ourselves.
    queue = deque([root])
    while queue:
        current = queue.popleft()
        if current.is_dir():
            for child in current.iterdir():
                queue.append(child)
        else:
            yield current


def is_cxx_source_file(path: Path) -> bool:
    return path.exists() and path.suffix in (
        ".cc",
        ".cp",
        ".cxx",
        ".cpp",
        ".CPP",
        ".c++",
        ".C",
        ".c",
    )


def is_cmake_file(path: Path) -> bool:
    return path.exists() and path.name.startswith("CMakeLists.txt")


def is_makefile_file(path: Path) -> bool:
    return path.exists() and path.name == "Makefile"


def is_makefile_target(makefile: Path, target: str) -> bool:
    if not is_makefile_file(makefile):
        return False
    proc = subprocess.run(
        ["make", "-qp"], cwd=makefile.parent, capture_output=True, text=True
    )
    for line in proc.stdout.splitlines():
        if line.strip().startswith(f"{target}:"):
            return True
    return False


def is_valid_target(target: str) -> bool:
    if target in (
        "all",
        "install",
        "depend",
        "test",
        "package",
        "package_source",
        "edit_cache",
        "rebuild_cache",
        "clean",
        "help",
        "ALL_BUILD",
        "ZERO_CHECK",
        "INSTALL",
        "RUN_TESTS",
        "PACKAGE",
    ):
        return False
    if target.endswith(".obj") or target.endswith(".i") or target.endswith(".s"):
        return False
    return True


def do_process(args: List[str | Path], **kwargs) -> subprocess.CompletedProcess:
    win_prefix: List[str | Path] = ["cmd", "/c"]
    cmd: List[str | Path] = args
    if sys.platform.startswith("win"):
        cmd = win_prefix + cmd
    return subprocess.run(cmd, **kwargs)


def run_cmd(
    cmd: List[str | Path],
    commands: Optional[List[List[str | Path]]] = None,
    stdout: Optional[List[str]] = None,
    stderr: Optional[List[str]] = None,
    working_directory: Optional[Path] = None,
):

    if working_directory is None:
        result = do_process(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
            timeout=LograderBasicConfig.DEFAULT_EXECUTABLE_TIMEOUT,
        )
    else:
        result = do_process(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
            timeout=LograderBasicConfig.DEFAULT_EXECUTABLE_TIMEOUT,
            cwd=working_directory,
        )

    if commands is not None:
        commands.append(cmd)
    if stdout is not None:
        stdout.append(result.stdout)
    if stderr is not None:
        stderr.append(result.stderr)
    return result



import sys
from pathlib import Path
from typing import Generator

import pytest

from lograder.tests import (
    TestCaseDict,
    TestCaseTemplate,
    TestRegistry,
    TSub,
    make_tests_from_files,
    make_tests_from_generator,
    make_tests_from_strs,
    make_tests_from_template,
)


@pytest.mark.description(
    "Testing the successful tests constructed from basic strings (using `make_tests_from_strs`)."
)
def test_basic_strs_success():
    TestRegistry.clear()

    num_tests: int = 10

    tests = make_tests_from_strs(
        names=[f"Test Case #{i}" for i in range(num_tests)],
        inputs=[f"{i}" for i in range(num_tests)],
        expected_outputs=[f"{i}" for i in range(num_tests)],
    )
    for i, test in enumerate(tests):
        assert test.get_expected_output() == f"{i}"


@pytest.mark.description(
    "Testing the unsuccessful tests constructed from basic strings (using `make_tests_from_strs`)."
)
def test_basic_strs_failure():
    TestRegistry.clear()

    num_tests: int = 10
    tests = make_tests_from_strs(
        names=[f"Test Case #{i}" for i in range(num_tests)],
        inputs=[f"{i}" for i in range(num_tests)],
        expected_outputs=[f"{i}" for i in range(num_tests)],
    )
    for i, test in enumerate(tests):
        assert test.get_expected_output() != "-1"


@pytest.mark.description("Testing to see if flags are passed correctly from tests.")
def test_flag_strs():
    TestRegistry.clear()

    num_tests: int = 10
    tests = make_tests_from_strs(
        names=[f"Test Case #{i}" for i in range(num_tests)],
        inputs=["" for _ in range(num_tests)],
        expected_outputs=[f"{i}" for i in range(num_tests)],
        flag_sets=[[f"{i}"] for i in range(num_tests)],
    )
    for i, test in enumerate(tests):
        if sys.platform.startswith("win"):
            test.set_target(["cmd", "/c", "echo"])
        else:
            test.set_target(["echo"])
        test.run()
        assert test.get_successful()


@pytest.mark.description("Testing to see if stdin is passed correctly from tests.")
def test_input_strs():
    TestRegistry.clear()

    num_tests: int = 10
    tests = make_tests_from_strs(
        names=[f"Test Case #{i}" for i in range(num_tests)],
        inputs=[f"{i}" for i in range(num_tests)],
        expected_outputs=[f"{i}" for i in range(num_tests)],
    )
    for i, test in enumerate(tests):
        if sys.platform.startswith("win"):
            test.set_target(["cmd", "/c", "more"])
        else:
            test.set_target(["cat"])
        test.run()
        assert test.get_successful()


@pytest.mark.description(
    "Testing the successful tests constructed from files (using `make_tests_from_files`)."
)
def test_basic_files_success():
    TestRegistry.clear()

    here = Path(__file__).parent
    tests = make_tests_from_files(
        names=["Test Case #0", "Test Case #1"],
        input_files=[
            here / "file_input" / "input-0.txt",
            here / "file_input" / "input-1.txt",
        ],
        expected_output_files=[
            here / "file_output" / "output-0.txt",
            here / "file_output" / "output-1.txt",
        ],
    )
    for i, test in enumerate(tests):
        assert test.get_expected_output() == f"{i}"


@pytest.mark.description(
    "Testing the unsuccessful tests constructed from files (using `make_tests_from_files`)."
)
def test_basic_files_failure():
    TestRegistry.clear()
    here = Path(__file__).parent
    tests = make_tests_from_files(
        names=["Test Case #0", "Test Case #1"],
        input_files=[
            here / "file_input" / "input-0.txt",
            here / "file_input" / "input-1.txt",
        ],
        expected_output_files=[
            here / "file_output" / "output-0.txt",
            here / "file_output" / "output-1.txt",
        ],
    )
    for i, test in enumerate(tests):
        assert test.get_expected_output() != "-1"


@pytest.mark.description(
    "Testing the successful tests constructed from template (using `make_tests_from_template`)."
)
def test_basic_templates_success():
    TestRegistry.clear()

    num_tests: int = 10
    here = Path(__file__).parent
    tests = make_tests_from_template(
        names=[f"Test Case #{i}" for i in range(num_tests)],
        template=TestCaseTemplate(
            input_template_str="{}",
            input_substitutions=[TSub(i) for i in range(num_tests)],
            expected_output_template_file=here / "file_template" / "template-0.txt",
            expected_output_substitutions=[TSub(i) for i in range(num_tests)],
        ),
    )
    for i, test in enumerate(tests):
        assert test.get_expected_output().strip() == f"{i}"


@pytest.mark.description(
    "Testing the unsuccessful tests constructed from template (using `make_tests_from_template`)."
)
def test_basic_templates_failure():
    TestRegistry.clear()

    num_tests: int = 10
    here = Path(__file__).parent
    tests = make_tests_from_template(
        names=[f"Test Case #{i}" for i in range(num_tests)],
        template=TestCaseTemplate(
            input_template_str="{}",
            input_substitutions=[TSub(i) for i in range(num_tests)],
            expected_output_template_file=here / "file_template" / "template-0.txt",
            expected_output_substitutions=[TSub(i) for i in range(num_tests)],
        ),
    )
    for i, test in enumerate(tests):
        assert test.get_expected_output() != "-1"


@pytest.mark.description(
    "Testing the successful tests constructed from function generator (using `make_tests_from_generator`)."
)
def test_basic_generator_success():
    TestRegistry.clear()

    num_tests: int = 10

    @make_tests_from_generator
    def make_tests() -> Generator[TestCaseDict, None, None]:
        for i in range(num_tests):
            case: TestCaseDict = {
                "name": f"Test Case #{i}",
                "input": f"{i}",
                "expected_output": f"{i}",
            }
            yield case

    for i, test in enumerate(TestRegistry.iterate()):
        assert test.get_expected_output() == f"{i}"


@pytest.mark.description(
    "Testing the unsuccessful tests constructed from function generator (using `make_tests_from_generator`)."
)
def test_basic_generator_failure():
    TestRegistry.clear()

    num_tests: int = 10

    @make_tests_from_generator
    def make_tests() -> Generator[TestCaseDict, None, None]:
        for i in range(num_tests):
            case: TestCaseDict = {
                "name": f"Test Case #{i}",
                "input": f"{i}",
                "expected_output": f"{i}",
            }
            yield case

    for i, test in enumerate(TestRegistry.iterate()):
        assert test.get_expected_output() != "-1"



from typing import Any, Dict, List, Optional

from pydantic import BaseModel, Field

from ..common import Status, TextFormat, Visibility


class TestCaseJSON(BaseModel):
    __test__: bool = False

    score: Optional[float]
    max_score: Optional[float]
    execution_time: Optional[float] = Field(default=None, exclude=True)
    status: Optional[Status] = None
    name: Optional[str]
    name_format: Optional[TextFormat] = "text"
    number: Optional[str] = None
    output: Optional[str]
    output_format: Optional[TextFormat] = "ansi"
    tags: List[str] = Field(default_factory=list)
    visibility: Optional[Visibility] = "hidden"
    extra_data: Optional[Dict[str, Any]] = Field(default_factory=dict)

    @property
    def is_scored(self) -> bool:
        return self.score is not None



import shlex
import subprocess
from pathlib import Path
from typing import List, Optional, Sequence

from ...dispatch.common.file_operations import do_process
from ...static import LograderBasicConfig
from ..common.exceptions import TestNotRunError, TestTargetNotSpecifiedError
from .analytics import (
    CallgrindSummary,
    ExecutionTimeSummary,
    ValgrindLeakSummary,
    ValgrindWarningSummary,
    callgrind,
    usr_time,
    valgrind,
)
from .interface import ExecutableTestInterface


class ExecutableOutputComparisonTest(ExecutableTestInterface):
    def __init__(
        self,
        name: str,
        input: str,
        expected_output: str,
        flags: Optional[Sequence[str | Path]] = None,
        weight: float = 1.0,
    ):
        if flags is None:
            flags = []

        self._name: str = name
        self._saved_executable: Optional[List[str | Path]] = None
        self._saved_flags: List[str | Path] = list(flags)

        self._cached_warnings: Optional[ValgrindWarningSummary] = None
        self._cached_leaks: Optional[ValgrindLeakSummary] = None
        self._cached_calls: Optional[List[CallgrindSummary]] = None
        self._cached_times: Optional[ExecutionTimeSummary] = None

        self._input: str = input
        self._expected_output: str = expected_output
        self._actual_output: Optional[str] = None
        self._error: Optional[str] = None

        self._weight: float = weight
        self._force_success: Optional[bool] = None  # I <3 ternary "boolean"

    def set_target(self, executable: List[str | Path]):
        self._saved_executable = executable

    def set_flags(self, flags: List[str | Path]):
        self._saved_flags = flags

    def override_output(self, stdout: str, stderr: str):
        self._error = stderr
        self._actual_output = stdout

    def force_successful(self) -> None:
        self._force_success = True

    def force_unsuccessful(self) -> None:
        self._force_success = False

    def run(
        self, wrap_args: bool = False, working_directory: Optional[Path] = None
    ) -> None:
        if self._force_success is not None:
            return  # no need to run test if its been overridden.

        cmd = self.get_cmd(wrap_args=wrap_args)
        if cmd is None:
            raise TestTargetNotSpecifiedError(self.get_name())

        self.set_target([cmd[0]])
        self.set_flags(cmd[1:])
        result = do_process(
            cmd,
            input=self.get_input(),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
            timeout=LograderBasicConfig.DEFAULT_EXECUTABLE_TIMEOUT,
            cwd=working_directory,
        )
        self._actual_output = result.stdout
        self._error = result.stderr

    def is_executed(self):
        if self._error is None or self._actual_output is None:
            return False
        return True

    def _assert_executed(self):
        if not self.is_executed():
            raise TestNotRunError(self.get_name())

    def get_error(self) -> str:
        self._assert_executed()
        assert self._error is not None  # mypy
        return self._error

    def get_warnings(self) -> Optional[ValgrindWarningSummary]:
        cmd = self.get_cmd()
        if cmd is None:
            raise TestNotRunError(self.get_name())
        if self._cached_warnings is None:
            self._cached_leaks, self._cached_warnings = valgrind(cmd, self.get_input())
        return self._cached_warnings

    def get_cmd(self, wrap_args: bool = False) -> Optional[List[Path | str]]:
        if self._saved_executable is None:
            return None

        flags: Sequence[str | Path] = self._saved_flags
        if wrap_args:
            flags = [
                f'ARGS="{shlex.join([str(arg.resolve()) if isinstance(arg, Path) else arg for arg in self._saved_flags])}"'
            ]
        return self._saved_executable + list(flags)

    def get_execution_time(self) -> Optional[ExecutionTimeSummary]:
        cmd = self.get_cmd()
        if cmd is None:
            raise TestNotRunError(self.get_name())
        if self._cached_times is None:
            self._cached_times = usr_time(cmd, self.get_input())
        return self._cached_times

    def get_calls(self) -> Optional[List[CallgrindSummary]]:
        cmd = self.get_cmd()
        if cmd is None:
            raise TestNotRunError(self.get_name())
        if self._cached_calls is None:
            self._cached_calls = callgrind(cmd, self.get_input())
        return self._cached_calls

    def get_leaks(self) -> Optional[ValgrindLeakSummary]:
        cmd = self.get_cmd()
        if cmd is None:
            raise TestNotRunError(self.get_name())
        if self._cached_warnings is None:
            self._cached_leaks, self._cached_warnings = valgrind(cmd, self.get_input())
        return self._cached_leaks

    def get_name(self) -> str:
        return self._name

    def get_successful(self) -> bool:
        if self._force_success is not None:
            return self._force_success
        actual_output = self.get_actual_output()
        assert actual_output is not None  # mypy
        return actual_output.strip() == self.get_expected_output().strip()

    def get_input(self) -> str:
        return self._input

    def get_expected_output(self) -> str:
        return self._expected_output

    def get_actual_output(self) -> Optional[str]:
        self._assert_executed()
        return self._actual_output

    def get_penalty(self) -> float:
        leaks = self.get_leaks()
        multiplier = 1.0
        if leaks is not None:
            if not leaks.is_safe:
                multiplier *= 0.8
        return multiplier

    def get_weight(self) -> float:
        return self._weight



import shlex
from datetime import timedelta
from os import PathLike
from typing import Optional, Sequence, Union

from colorama import Fore

from ...dispatch.common.types import (
    AssignmentMetadata,
    BuilderOutput,
    PreprocessorOutput,
)
from ...static import LograderBasicConfig
from ...tests.test.analytics import (
    CallgrindSummary,
    ExecutionTimeSummary,
    ValgrindLeakSummary,
    ValgrindWarningSummary,
)
from ...tests.test.interface import ExecutableTestInterface
from .format_templates import ContextRenderer, ProcessStep, RendererInterface
from .interfaces import (
    BuildOutputFormatterInterface,
    ExecutableTestFormatterInterface,
    ExecutionTimeSummaryFormatterInterface,
    MetadataFormatterInterface,
    PreprocessorOutputFormatterInterface,
    RuntimeSummaryFormatterInterface,
    ValgrindLeakSummaryFormatterInterface,
    ValgrindWarningSummaryFormatterInterface,
)


def format_timedelta(td: timedelta) -> str:
    if td < timedelta(0):
        return "0s"
    total_seconds = int(td.total_seconds())
    milliseconds = int(td.microseconds / 1000)

    weeks, remainder = divmod(total_seconds, 7 * 24 * 3600)
    days, remainder = divmod(remainder, 24 * 3600)
    hours, remainder = divmod(remainder, 3600)
    minutes, seconds = divmod(remainder, 60)

    parts = []
    if weeks:
        parts.append(f"{weeks}w")
    if days:
        parts.append(f"{days}d")
    if hours:
        parts.append(f"{hours}h")
    if minutes:
        parts.append(f"{minutes}m")
    if seconds:
        parts.append(f"{seconds}s")
    if milliseconds:
        parts.append(f"{milliseconds}ms")

    return " ".join(parts) if parts else "0s"


class CLIStep(ProcessStep):
    def __init__(self, command: Sequence[Union[str, bytes, PathLike]]):
        self.command = shlex.join([str(comm) for comm in command])

    def render(self):
        return f"Ran `{Fore.MAGENTA}{self.command}{Fore.RESET}` in CLI."


class DefaultMetadataFormatter(MetadataFormatterInterface):
    def format(self, assignment_metadata: AssignmentMetadata):
        return (
            f"`{Fore.MAGENTA}{assignment_metadata.assignment_name}{Fore.RESET}` made by {Fore.MAGENTA}{f'{Fore.RESET}, {Fore.MAGENTA}'.join(assignment_metadata.assignment_authors)}{Fore.RESET}.\n"
            f"Submission date: {Fore.MAGENTA}{assignment_metadata.assignment_submit_date.strftime("%Y-%m-%d %H:%M:%S.%f")}{Fore.RESET}.\n"
            f"Due date: {Fore.MAGENTA}{assignment_metadata.assignment_due_date.strftime("%Y-%m-%d %H:%M:%S.%f")}{Fore.RESET}.\n"
            f"Time left: {Fore.MAGENTA}{format_timedelta(assignment_metadata.assignment_due_date - assignment_metadata.assignment_submit_date)}{Fore.RESET}.\n\n"
            f"Assignment graded with `{Fore.GREEN}{assignment_metadata.library_name}{Fore.RESET}` (version {Fore.GREEN}{assignment_metadata.library_version}{Fore.RESET}), made by {Fore.GREEN}{f'{Fore.RESET}, {Fore.GREEN}'.join(assignment_metadata.library_authors)}.{Fore.RESET}\n\n"
        )


class DefaultPreprocessorContext(
    ContextRenderer,
    prefix=f"<{Fore.CYAN}BEGIN PREPROCESSOR{Fore.RESET}>\n",
    suffix=f"\n<{Fore.CYAN}END PREPROCESSOR{Fore.RESET}>",
    empty=f"<{Fore.CYAN}NO PREPROCESSOR PRESENT{Fore.RESET}>",
):
    pass


class DefaultBuilderContext(
    ContextRenderer,
    prefix=f"<{Fore.YELLOW}BEGIN BUILDER{Fore.RESET}>\n",
    suffix=f"\n<{Fore.YELLOW}END BUILDER{Fore.RESET}>",
    empty=f"<{Fore.YELLOW}NO BUILDER PRESENT{Fore.RESET}>",
):
    pass


class DefaultSTDOUTContext(
    ContextRenderer,
    prefix=f"<{Fore.BLUE}BEGIN STDOUT{Fore.RESET}>\n",
    suffix=f"\n<{Fore.BLUE}END STDOUT{Fore.RESET}>",
    empty=f"<{Fore.BLUE}EMPTY STDOUT{Fore.RESET}>",
):
    pass


class DefaultExpectedSTDOUTContext(
    ContextRenderer,
    prefix=f"<{Fore.BLUE}BEGIN EXPECTED STDOUT{Fore.RESET}>\n",
    suffix=f"\n<{Fore.BLUE}END EXPECTED STDOUT{Fore.RESET}>",
    empty=f"<{Fore.BLUE}EMPTY EXPECTED STDOUT{Fore.RESET}>",
):
    def render(self) -> str:
        context = super().render()
        comp_repr = (
            f"({Fore.BLUE}RAW EXPECTED STDOUT{Fore.RESET}): {repr(self.get_content())}"
        )
        return f"{context}\n{comp_repr}"


class DefaultActualSTDOUTContext(
    ContextRenderer,
    prefix=f"<{Fore.LIGHTMAGENTA_EX}BEGIN ACTUAL STDOUT{Fore.RESET}>\n",
    suffix=f"\n<{Fore.LIGHTMAGENTA_EX}END ACTUAL STDOUT{Fore.RESET}>",
    empty=f"<{Fore.LIGHTMAGENTA_EX}EMPTY ACTUAL STDOUT{Fore.RESET}>",
):
    def __init__(self, content: Union[RendererInterface, str, None]):
        if content is None:
            content = ""
        if isinstance(content, str):
            content = content.strip()
        super().__init__(content)

    def render(self) -> str:
        context = super().render()
        comp_repr = f"({Fore.LIGHTMAGENTA_EX}RAW ACTUAL STDOUT{Fore.RESET}): {repr(self.get_content())}"
        return f"{context}\n{comp_repr}"


class DefaultSTDERRContext(
    ContextRenderer,
    prefix=f"<{Fore.RED}BEGIN STDERR{Fore.RESET}>\n",
    suffix=f"\n<{Fore.RED}END STDERR{Fore.RESET}>",
    empty=f"<{Fore.RED}EMPTY STDERR{Fore.RESET}>",
):
    pass


class DefaultSTDINContext(
    ContextRenderer,
    prefix=f"<{Fore.MAGENTA}BEGIN STDIN{Fore.RESET}>\n",
    suffix=f"\n<{Fore.MAGENTA}END STDIN{Fore.RESET}>",
    empty=f"<{Fore.MAGENTA}EMPTY STDIN{Fore.RESET}>",
):
    pass


class DefaultValgrindLeakSummaryFormatter(ValgrindLeakSummaryFormatterInterface):
    def format(self, leak_summary: Optional[ValgrindLeakSummary]) -> str:
        if leak_summary is None:
            return f"<{Fore.YELLOW}VALGRIND LEAK SUMMARY DISABLED{Fore.RESET}>"
        def_lost_color = (
            Fore.LIGHTGREEN_EX if leak_summary.definitely_lost.is_safe else Fore.RED
        )
        ind_lost_color = (
            Fore.LIGHTGREEN_EX if leak_summary.indirectly_lost.is_safe else Fore.RED
        )
        pos_lost_color = (
            Fore.LIGHTGREEN_EX if leak_summary.possibly_lost.is_safe else Fore.RED
        )
        return (
            f"{Fore.LIGHTGREEN_EX}VALGRIND LEAK SUMMARY{Fore.RESET}:\n"
            f"  {Fore.LIGHTBLUE_EX}*{Fore.RESET} {def_lost_color}{leak_summary.definitely_lost.bytes}{Fore.RESET} bytes, {def_lost_color}{leak_summary.definitely_lost.blocks}{Fore.RESET} blocks {def_lost_color}definitely lost{Fore.RESET}.\n"
            f"  {Fore.LIGHTBLUE_EX}*{Fore.RESET} {ind_lost_color}{leak_summary.indirectly_lost.bytes}{Fore.RESET} bytes, {ind_lost_color}{leak_summary.indirectly_lost.blocks}{Fore.RESET} blocks {ind_lost_color}indirectly lost{Fore.RESET}.\n"
            f"  {Fore.LIGHTBLUE_EX}*{Fore.RESET} {pos_lost_color}{leak_summary.possibly_lost.bytes}{Fore.RESET} bytes, {pos_lost_color}{leak_summary.possibly_lost.blocks}{Fore.RESET} blocks {pos_lost_color}possibly lost{Fore.RESET}.\n"
            f"  {Fore.LIGHTBLUE_EX}*{Fore.RESET} {Fore.LIGHTGREEN_EX}{leak_summary.still_reachable.bytes}{Fore.RESET} bytes, {Fore.LIGHTGREEN_EX}{leak_summary.still_reachable.blocks}{Fore.RESET} blocks {Fore.LIGHTGREEN_EX}still reachable{Fore.RESET}."
        )


class DefaultValgrindWarningSummaryFormatter(ValgrindWarningSummaryFormatterInterface):
    def format(self, warning_summary: Optional[ValgrindWarningSummary]) -> str:
        if warning_summary is None:
            return f"<{Fore.YELLOW}VALGRIND WARNING SUMMARY DISABLED{Fore.RESET}>"
        warning = warning_summary.model_dump()
        output = [f"{Fore.LIGHTGREEN_EX}VALGRIND WARNING SUMMARY{Fore.RESET}:"]
        output += [
            f"  {Fore.LIGHTBLUE_EX}*{Fore.RESET} {Fore.LIGHTGREEN_EX if v == 0 else Fore.RED}{v}{Fore.RESET} `{Fore.LIGHTRED_EX}{k.replace('_', ' ').upper()}{Fore.RESET}` warnings encountered."
            for k, v in warning.items()
        ]
        return "\n".join(output)


class DefaultExecutionTimeSummaryFormatter(ExecutionTimeSummaryFormatterInterface):
    def format(
        self,
        callgrind_summary: Optional[Sequence[CallgrindSummary]],
        execution_time_summary: Optional[ExecutionTimeSummary],
    ) -> str:
        if callgrind_summary is None or execution_time_summary is None:
            return f"<{Fore.YELLOW}PERFORMANCE SUMMARY DISABLED{Fore.RESET}>"
        total_time = execution_time_summary.total_cpu_time
        output = [
            f"{Fore.LIGHTGREEN_EX}PERFORMANCE SUMMARY{Fore.RESET}:",
            "Time elapsed on CPU:",
            f"  {Fore.LIGHTBLUE_EX}*{Fore.RESET} Total: {Fore.LIGHTGREEN_EX}{format_timedelta(timedelta(seconds=total_time))}{Fore.RESET}",
            f"  {Fore.LIGHTBLUE_EX}*{Fore.RESET} User-initiated tasks: {Fore.LIGHTGREEN_EX}{format_timedelta(timedelta(seconds=execution_time_summary.user_cpu_time))}{Fore.RESET}",
            f"  {Fore.LIGHTBLUE_EX}*{Fore.RESET} System-initiated tasks: {Fore.LIGHTGREEN_EX}{format_timedelta(timedelta(seconds=execution_time_summary.system_cpu_time))}{Fore.RESET}",
            f"{Fore.LIGHTGREEN_EX}{execution_time_summary.percent_cpu_utilization:.2f}%{Fore.RESET} CPU Usage",
            f"{Fore.LIGHTGREEN_EX}{execution_time_summary.peak_physical_memory_usage:.2f} KB{Fore.RESET} Peak Memory Usage",
            f"Read from disk {Fore.LIGHTGREEN_EX}{execution_time_summary.num_disk_reads}{Fore.RESET} times",
            f"Wrote to disk {Fore.LIGHTGREEN_EX}{execution_time_summary.num_disk_writes}{Fore.RESET} times",
            f"{Fore.LIGHTGREEN_EX}{execution_time_summary.num_major_page_faults}{Fore.RESET} pages fetched from disk (major page-faults).",
            f"{Fore.LIGHTGREEN_EX}{execution_time_summary.num_minor_page_faults}{Fore.RESET} pages mapped from RAM (minor page-faults).",
            f"{Fore.LIGHTGREEN_EX}{execution_time_summary.num_pages_swapped_to_disk}{Fore.RESET} pages swapped to disk.\n",
            f"{Fore.LIGHTGREEN_EX}FUNCTION CALL BREAKDOWN{Fore.RESET}:",
        ]
        function_text = []

        total_instructions = sum([call.cost for call in callgrind_summary])
        for call_summary in callgrind_summary:
            function_text.append(
                f"  {Fore.LIGHTBLUE_EX}*{Fore.RESET} {Fore.LIGHTGREEN_EX}{call_summary.percent:.2f}%{Fore.RESET} ({Fore.LIGHTGREEN_EX}{call_summary.cost}{Fore.RESET} inst., {Fore.LIGHTBLUE_EX}~{Fore.RESET}{Fore.LIGHTGREEN_EX}{format_timedelta(timedelta(seconds=total_time * call_summary.cost / total_instructions))}{Fore.RESET}): {Fore.LIGHTMAGENTA_EX}{call_summary.function}{Fore.RESET} from {Fore.MAGENTA}{call_summary.file}{Fore.RESET} {'in ' + Fore.BLUE + call_summary.shared_object + Fore.RESET if call_summary.shared_object is not None else ''}"
            )
        if not function_text:
            function_text = [
                f"  {Fore.LIGHTBLUE_EX}*{Fore.RESET} <{Fore.RED}EMPTY?{Fore.RESET}>"
            ]

        return "\n".join(output + function_text)


class DefaultPreprocessorOutputFormatter(PreprocessorOutputFormatterInterface):
    def format(self, preprocessor_output: PreprocessorOutput) -> str:
        output = [
            "\n".join(
                [
                    line
                    for line in (
                        f"{CLIStep(command).render()}",
                        f"{DefaultSTDOUTContext(cout.strip()).render()}",
                        f"{DefaultSTDERRContext(cerr.strip()).render()}",
                    )
                    if line
                ]
            )
            for command, cout, cerr in zip(
                preprocessor_output.commands,
                preprocessor_output.stdout,
                preprocessor_output.stderr,
            )
        ]
        return DefaultPreprocessorContext("\n\n".join(output).strip()).render()


class DefaultBuildOutputFormatter(BuildOutputFormatterInterface):
    def format(self, build_output: BuilderOutput) -> str:
        output = [
            f"Detected build types of `{Fore.MAGENTA}{build_output.project_type}{Fore.RESET}`."
        ]
        output += [
            "\n".join(
                [
                    line
                    for line in (
                        f"{CLIStep(command).render()}",
                        f"{DefaultSTDOUTContext(cout.strip()).render()}",
                        f"{DefaultSTDERRContext(cerr.strip()).render()}",
                    )
                    if line
                ]
            )
            for command, cout, cerr in zip(
                build_output.commands, build_output.stdout, build_output.stderr
            )
        ]
        return DefaultBuilderContext("\n\n".join(output).strip()).render()


class DefaultRuntimeSummaryFormatter(RuntimeSummaryFormatterInterface):
    def format(self, test_cases: Sequence[ExecutableTestInterface]) -> str:
        total_cpu_time = sum(
            (et.total_cpu_time if (et := tc.get_execution_time()) is not None else 0.0)
            for tc in test_cases
        )
        user_cpu_time = sum(
            (et.user_cpu_time if (et := tc.get_execution_time()) is not None else 0.0)
            for tc in test_cases
        )
        system_cpu_time = sum(
            (et.system_cpu_time if (et := tc.get_execution_time()) is not None else 0.0)
            for tc in test_cases
        )
        call_lists: list[list[CallgrindSummary]] = [
            calls
            for test_case in test_cases
            if (calls := test_case.get_calls()) is not None
        ]
        total_instructions: int = sum(
            [sum(call.cost for call in calls) for calls in call_lists]
        )
        num_tests = len(test_cases)
        num_successful_tests = sum(
            [test_case.get_successful() for test_case in test_cases]
        )
        if num_successful_tests < 0.5 * num_tests:
            color = Fore.RED
        elif num_successful_tests < 0.75 * num_tests:
            color = Fore.LIGHTRED_EX
        elif num_successful_tests < 0.95 * num_tests:
            color = Fore.YELLOW
        elif num_successful_tests <= num_tests:
            color = Fore.GREEN
        else:
            color = Fore.CYAN
        return (
            f"{color}{num_successful_tests}{Fore.RESET}/{num_tests} Tests Passed.\n"
            f"{Fore.LIGHTMAGENTA_EX}Total Time Elapsed on CPU{Fore.RESET}:\n"
            f"  {Fore.LIGHTBLUE_EX}*{Fore.RESET} Total: {Fore.LIGHTGREEN_EX}{format_timedelta(timedelta(seconds=total_cpu_time))}{Fore.RESET}\n"
            f"  {Fore.LIGHTBLUE_EX}*{Fore.RESET} User-initiated tasks: {Fore.LIGHTGREEN_EX}{format_timedelta(timedelta(seconds=user_cpu_time))}{Fore.RESET}\n"
            f"  {Fore.LIGHTBLUE_EX}*{Fore.RESET} System-initiated tasks: {Fore.LIGHTGREEN_EX}{format_timedelta(timedelta(seconds=system_cpu_time))}{Fore.RESET}\n"
            f"{Fore.LIGHTMAGENTA_EX}Total Number of Instructions Run{Fore.RESET}: {Fore.LIGHTGREEN_EX}{total_instructions}{Fore.RESET}."
        )


class DefaultExecutableTestCaseFormatter(ExecutableTestFormatterInterface):
    def format(self, test_case: ExecutableTestInterface) -> str:
        if not test_case.get_successful():
            title_text = f"{Fore.RED}Test `{test_case.get_name()}` failed!{Fore.RESET}"
        elif test_case.get_penalty() < 1.0:
            title_text = f"{Fore.YELLOW}Test `{test_case.get_name()}` partially passed, but was penalized!{Fore.RESET}"
        elif test_case.get_penalty() == 1.0:
            title_text = (
                f"{Fore.GREEN}Test `{test_case.get_name()}` passed!{Fore.RESET}"
            )
        else:
            title_text = f"{Fore.CYAN}Test `{test_case.get_name()}` passed with flying colors (bonus points)!{Fore.RESET}!"
        output = [
            title_text,
            LograderBasicConfig.DEFAULT_TOPIC_BREAK,
            DefaultSTDINContext(test_case.get_input()).render(),
            LograderBasicConfig.DEFAULT_TOPIC_BREAK,
            DefaultExpectedSTDOUTContext(
                test_case.get_expected_output().strip()
            ).render(),
            LograderBasicConfig.DEFAULT_TOPIC_BREAK,
            DefaultActualSTDOUTContext(test_case.get_actual_output()).render(),
            LograderBasicConfig.DEFAULT_TOPIC_BREAK,
            DefaultSTDERRContext(test_case.get_error().strip()).render(),
            LograderBasicConfig.DEFAULT_TOPIC_BREAK,
            DefaultValgrindLeakSummaryFormatter().format(test_case.get_leaks()),
            LograderBasicConfig.DEFAULT_TOPIC_BREAK,
            DefaultValgrindWarningSummaryFormatter().format(test_case.get_warnings()),
            LograderBasicConfig.DEFAULT_TOPIC_BREAK,
            DefaultExecutionTimeSummaryFormatter().format(
                test_case.get_calls(), test_case.get_execution_time()
            ),
        ]
        return "".join(output)



from ._core_exceptions import LograderError
from .tests.exceptions import (
    LograderBuildError,
    LograderTestError,
    LograderValidationError,
    MismatchedSequenceLengthError,
)

__all__ = [
    "LograderError",
    "LograderValidationError",
    "LograderBuildError",
    "LograderTestError",
    "MismatchedSequenceLengthError",
]



from .common import FilePath
from .file import make_tests_from_files
from .generator import (
    FlaggedTestCaseProtocol,
    FlaggedWeightedTestCaseProtocol,
    TestCaseDict,
    TestCaseProtocol,
    WeightedTestCaseProtocol,
    make_tests_from_generator,
)
from .registry import TestRegistry
from .simple import make_tests_from_strs
from .template import (
    TemplateSubstitution,
    TestCaseTemplate,
    TSub,
    make_tests_from_template,
)
from .test import ExecutableOutputComparisonTest

__all__ = [
    "make_tests_from_strs",
    "FilePath",
    "make_tests_from_files",
    "TestCaseProtocol",
    "WeightedTestCaseProtocol",
    "FlaggedTestCaseProtocol",
    "FlaggedWeightedTestCaseProtocol",
    "TestCaseDict",
    "make_tests_from_generator",
    "TemplateSubstitution",
    "TSub",
    "TestCaseTemplate",
    "make_tests_from_template",
    "ExecutableOutputComparisonTest",
    "TestRegistry",
]



from ._core_exceptions import LograderBuildError, LograderError, LograderTestError
from .common.exceptions import LograderValidationError, MismatchedSequenceLengthError

__all__ = [
    "LograderError",
    "LograderTestError",
    "LograderBuildError",
    "MismatchedSequenceLengthError",
    "LograderValidationError",
]



from .cli_builder import CLIBuilder
from .executable_runner import ExecutableRunner
from .trivial import TrivialBuilder, TrivialPreprocessor

__all__ = [
    "ExecutableRunner",
    "CLIBuilder",
    "TrivialPreprocessor",
    "TrivialBuilder",
]



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, List, Optional, Sequence

from ...dispatch.common.types import (
    AssignmentMetadata,
    BuilderOutput,
    PreprocessorOutput,
)
from ...tests.test.analytics import (
    CallgrindSummary,
    ExecutionTimeSummary,
    ValgrindLeakSummary,
    ValgrindWarningSummary,
)

if TYPE_CHECKING:
    from ...tests.test.interface import ExecutableTestInterface


class MetadataFormatterInterface(ABC):
    @abstractmethod
    def format(self, assignment_metadata: AssignmentMetadata) -> str:
        pass


class PreprocessorOutputFormatterInterface(ABC):
    @abstractmethod
    def format(self, preprocessor_output: PreprocessorOutput) -> str:
        pass


class BuildOutputFormatterInterface(ABC):
    @abstractmethod
    def format(self, build_output: BuilderOutput) -> str:
        pass


class RuntimeSummaryFormatterInterface(ABC):
    @abstractmethod
    def format(self, test_cases: Sequence[ExecutableTestInterface]) -> str:
        pass


class ValgrindLeakSummaryFormatterInterface(ABC):
    @abstractmethod
    def format(self, leak_summary: Optional[ValgrindLeakSummary]) -> str:
        pass


class ValgrindWarningSummaryFormatterInterface(ABC):
    @abstractmethod
    def format(self, warning_summary: Optional[ValgrindWarningSummary]) -> str:
        pass


class ExecutionTimeSummaryFormatterInterface(ABC):
    @abstractmethod
    def format(
        self,
        callgrind_summary: Optional[List[CallgrindSummary]],
        execution_time_summary: Optional[ExecutionTimeSummary],
    ) -> str:
        pass


class ExecutableTestFormatterInterface(ABC):
    @abstractmethod
    def format(self, test_case: ExecutableTestInterface) -> str:
        pass



from __future__ import annotations

from abc import ABC, abstractmethod
from pathlib import Path
from typing import List, Optional

from .analytics import (
    CallgrindSummary,
    ExecutionTimeSummary,
    ValgrindLeakSummary,
    ValgrindWarningSummary,
)


class TestInterface(ABC):
    __test__: bool = False

    @abstractmethod
    def run(
        self, wrap_args: bool = False, working_directory: Optional[Path] = None
    ) -> None:
        pass

    @abstractmethod
    def get_name(self) -> str:
        pass

    @abstractmethod
    def get_successful(self) -> bool:
        pass

    @abstractmethod
    def get_weight(self) -> float:
        pass

    @abstractmethod
    def force_successful(self) -> None:
        pass

    @abstractmethod
    def force_unsuccessful(self) -> None:
        pass

    def get_penalty(self) -> float:
        return 1.0


class ExecutableTestInterface(TestInterface, ABC):
    @abstractmethod
    def set_target(self, target: List[str | Path]):
        pass

    @abstractmethod
    def is_executed(self) -> bool:
        pass

    @abstractmethod
    def get_expected_output(self) -> str:
        pass

    @abstractmethod
    def get_error(self) -> str:
        pass

    @abstractmethod
    def get_input(self) -> str:
        pass

    @abstractmethod
    def get_actual_output(self) -> Optional[str]:
        pass

    @abstractmethod
    def override_output(self, stdout: str, stderr: str):
        pass

    def get_warnings(self) -> Optional[ValgrindWarningSummary]:
        return None

    def get_execution_time(self) -> Optional[ExecutionTimeSummary]:
        return None

    def get_calls(self) -> Optional[List[CallgrindSummary]]:
        return None

    def get_leaks(self) -> Optional[ValgrindLeakSummary]:
        return None



from .test_maker import make_tests_from_generator
from .types import (
    FlaggedTestCaseProtocol,
    FlaggedWeightedTestCaseProtocol,
    TestCaseDict,
    TestCaseProtocol,
    WeightedTestCaseProtocol,
)

__all__ = [
    "TestCaseProtocol",
    "WeightedTestCaseProtocol",
    "FlaggedWeightedTestCaseProtocol",
    "FlaggedTestCaseProtocol",
    "TestCaseDict",
    "make_tests_from_generator",
]



from typing import Dict, Generator, List, Sequence

from ..test import TestInterface


class Registry[T]:
    _registry: Dict[str, T] = {}
    _container: List[T] = []

    @classmethod
    def clear(cls):
        cls._registry = {}
        cls._container = []

    @classmethod
    def register(cls, key: str, value: T) -> None:
        cls._registry[key] = value

    @classmethod
    def deregister(cls, key: str) -> None:
        del cls._registry[key]

    @classmethod
    def get(cls, key: str) -> T:
        return cls._registry[key]

    @classmethod
    def add(cls, value: T) -> None:
        cls._container.append(value)

    @classmethod
    def extend(cls, values: Sequence[T]) -> None:
        cls._container.extend(values)

    @classmethod
    def remove(cls, value: T) -> None:
        cls._container.remove(value)

    @classmethod
    def iterate(cls) -> Generator[T, None, None]:
        yield from cls._container


TestRegistry = Registry[TestInterface]



from pathlib import Path
from typing import List

from .._core_exceptions import LograderStudentBuildError


class CxxSourceBuildError(LograderStudentBuildError):
    def __init__(self, sources: List[Path]):
        super().__init__(
            f"Was unable to compile C++ sources; found source files: {', '.join([str(path.resolve()) for path in sources])}."
        )


class CMakeListsNotFoundError(LograderStudentBuildError):
    def __init__(self):
        super().__init__("Could not find a `CMakeLists.txt` anywhere in the project.")


class CMakeTargetNotFoundError(LograderStudentBuildError):
    def __init__(self, targets: List[str], cmake_path: Path):
        super().__init__(
            f"Could not find a valid cmake target anywhere in `{cmake_path.resolve()}`. The targets found were: [{', '.join(targets)}]."
        )


class CMakeExecutableNotFoundError(LograderStudentBuildError):
    def __init__(self, cmake_path: Path):
        super().__init__(
            f"Could not find a valid cmake executable output path anywhere in `{cmake_path.resolve()}`."
        )


class MakefileNotFoundError(LograderStudentBuildError):
    def __init__(self):
        super().__init__("Could not find a `Makefile` anywhere in the project.")


class MakefileRunNotFoundError(LograderStudentBuildError):
    def __init__(self, makefile_path: Path):
        super().__init__(
            f"Could not find an `run` entrypoint in `{makefile_path.resolve()}` (i.e. the command `make run ARGS=<test-args>` is used to run the project)."
        )



import os
import re
import subprocess
import sys
from pathlib import Path
from typing import Any, List, Optional

from pydantic import BaseModel, Field

from ...common.utils import random_name
from ...static import LograderBasicConfig


class LossEntry(BaseModel):
    bytes: int = Field(default=0)
    blocks: int = Field(default=0)

    @property
    def is_safe(self) -> bool:
        return not self.bytes and not self.blocks


class ValgrindLeakSummary(BaseModel):
    definitely_lost: LossEntry = Field(default_factory=LossEntry)
    indirectly_lost: LossEntry = Field(default_factory=LossEntry)
    possibly_lost: LossEntry = Field(default_factory=LossEntry)
    still_reachable: LossEntry = Field(default_factory=LossEntry)

    @property
    def is_safe(self) -> bool:
        return (
            self.definitely_lost.is_safe
            and self.indirectly_lost.is_safe
            and self.possibly_lost.is_safe
        )


class ValgrindWarningSummary(BaseModel):
    invalid_read: int = Field(default=0)
    invalid_write: int = Field(default=0)
    invalid_free: int = Field(default=0)
    mismatched_free: int = Field(default=0)
    uninitialized_value: int = Field(default=0)
    conditional_jump: int = Field(default=0)
    syscall_param: int = Field(default=0)
    overlap: int = Field(default=0)
    other: int = Field(default=0)  # fallback bucket

    @property
    def is_safe(self) -> bool:
        return (
            not self.invalid_read
            and not self.invalid_write
            and not self.invalid_free
            and not self.mismatched_free
            and not self.uninitialized_value
            and not self.conditional_jump
            and not self.syscall_param
            and not self.overlap
            and not self.other
        )


class ValgrindOutput:
    LEAK_REGEX = re.compile(
        r"(\d+)\s+bytes in\s+(\d+)\s+blocks are (definitely lost|indirectly lost|possibly lost|still reachable)"
    )
    WARNING_PATTERNS = {
        "invalid_read": re.compile(r"Invalid read"),
        "invalid_write": re.compile(r"Invalid write"),
        "invalid_free": re.compile(r"Invalid free"),
        "mismatched_free": re.compile(r"Mismatched free"),
        "uninitialized_value": re.compile(r"uninitialised value"),
        "conditional_jump": re.compile(
            r"Conditional jump or move depends on uninitialised value"
        ),
        "syscall_param": re.compile(r"Syscall param"),
        "overlap": re.compile(r"overlap"),
    }

    def __init__(self, stderr: str):
        self._stderr: str = stderr
        self._warnings: ValgrindWarningSummary = ValgrindWarningSummary()
        self._leaks: ValgrindLeakSummary = ValgrindLeakSummary()
        self.parse_stderr()

    @classmethod
    def parse_valgrind_log(
        cls, stderr: str
    ) -> tuple[ValgrindLeakSummary, ValgrindWarningSummary]:
        # Init structures
        leaks: ValgrindLeakSummary = ValgrindLeakSummary()
        warnings: ValgrindWarningSummary = ValgrindWarningSummary()
        warnings.other = 0

        for line in stderr.split("\n"):
            # --- Leak parsing ---
            leak_match = cls.LEAK_REGEX.search(line)
            if leak_match:
                bytes_count = int(leak_match.group(1).replace(",", ""))
                blocks_count = int(leak_match.group(2).replace(",", ""))
                kind = leak_match.group(3).replace(" ", "_")  # normalize to dict key
                prev_bytes, prev_blocks = getattr(leaks, kind)
                setattr(
                    leaks,
                    kind,
                    LossEntry(
                        bytes=prev_bytes + bytes_count,
                        blocks=prev_blocks + blocks_count,
                    ),
                )
                continue

            # --- Warning parsing ---
            matched = False
            for key, pattern in cls.WARNING_PATTERNS.items():
                if pattern.search(line):
                    setattr(warnings, key, getattr(warnings, key) + 1)
                    matched = True
                    break
            if not matched and "==" in line and "==" in line.strip():
                # heuristic: unknown Valgrind warning line
                if not any(x in line for x in ["lost", "reachable"]):
                    warnings.other += 1

        return leaks, warnings

    def parse_stderr(self):
        self._leaks, self._warnings = self.parse_valgrind_log(self._stderr)

    def get_leaks(self) -> ValgrindLeakSummary:
        return self._leaks

    def get_warnings(self) -> ValgrindWarningSummary:
        return self._warnings


class CallgrindSummary(BaseModel):
    cost: int
    percent: float
    file: str
    function: str
    shared_object: Optional[str]


class CallgrindOutput:
    LINE_REGEX = re.compile(
        r"^\s*(?P<cost>[\d,]+)\s+\((?P<percent>[\d\.]+)%\)\s+"
        r"(?P<file>[^:]+):(?P<function>[^\[]+)"
        r"(?:\s+\[(?P<so>[^\]]+)\])?"
    )

    def __init__(self, stdout: str):
        self._stdout: str = stdout
        self._calls: List[CallgrindSummary] = []
        self.parse_stdout()

    @classmethod
    def parse_callgrind_annotate(cls, stdout: str) -> List[CallgrindSummary]:
        results: List[CallgrindSummary] = []
        for line in stdout.split("\n"):
            m = cls.LINE_REGEX.match(line)
            if not m:
                continue
            results.append(
                CallgrindSummary(
                    cost=int(m.group("cost").replace(",", "")),
                    percent=float(m.group("percent")),
                    file=m.group("file").strip(),
                    function=m.group("function").strip(),
                    shared_object=m.group("so"),
                )
            )
        return results

    def parse_stdout(self):
        self._calls = self.parse_callgrind_annotate(self._stdout)

    def get_calls(self):
        return self._calls

    def get_instruction_count(self):
        return sum([call.cost for call in self.get_calls()])


class ExecutionTimeSummary(BaseModel):
    user_cpu_time: float = Field(default=0)
    system_cpu_time: float = Field(default=0)
    total_cpu_time: float = Field(default=0)
    percent_cpu_utilization: float = Field(default=0)
    peak_physical_memory_usage: float = Field(default=0)
    num_disk_reads: int = Field(default=0)
    num_disk_writes: int = Field(default=0)
    num_major_page_faults: int = Field(default=0)
    num_minor_page_faults: int = Field(default=0)
    num_pages_swapped_to_disk: int = Field(default=0)


class TimeOutput:
    def __init__(self, stderr: str):
        self._stderr: str = stderr
        self._time: ExecutionTimeSummary = ExecutionTimeSummary()
        self.parse_stderr()

    @staticmethod
    def parse_usr_time(stderr: str) -> ExecutionTimeSummary:
        stats: dict[str, Any] = {}
        for line in stderr.splitlines():
            if "=" in line:
                k, v = line.split("=", 1)
                stats[k.strip()] = v.strip().replace("%", "")
        return ExecutionTimeSummary(**stats)

    def parse_stderr(self):
        self._time = self.parse_usr_time(self._stderr)

    def get_time(self) -> ExecutionTimeSummary:
        return self._time


def valgrind(
    cmd: List[str | Path], stdin: Optional[str] = None
) -> tuple[ValgrindLeakSummary, ValgrindWarningSummary]:

    if sys.platform.startswith("win"):
        return ValgrindLeakSummary(), ValgrindWarningSummary()

    valgrind_file = f"valgrind-{random_name()}.log"
    with open(os.devnull, "w") as devnull:
        result = subprocess.run(
            [
                "valgrind",
                "--leak-check=full",
                "--show-leak-kinds=all",
                f"--log-file={valgrind_file}",
            ]
            + cmd,
            input=stdin,
            stdout=devnull,
            stderr=devnull,
            text=True,
            timeout=LograderBasicConfig.DEFAULT_EXECUTABLE_TIMEOUT,
        )

    if result.returncode != 0:
        if Path(valgrind_file).is_file():
            os.remove(valgrind_file)
        return ValgrindLeakSummary(), ValgrindWarningSummary()

    with open(valgrind_file, "r", encoding="utf-8", errors="ignore") as f:
        valgrind_log = f.read()
    valgrind_output = ValgrindOutput(valgrind_log)
    os.remove(valgrind_file)

    return valgrind_output.get_leaks(), valgrind_output.get_warnings()


def callgrind(
    cmd: List[Path | str], stdin: Optional[str] = None
) -> List[CallgrindSummary]:

    if sys.platform.startswith("win"):
        return []

    callgrind_file = f"callgrind-{random_name()}.out"
    annotate_file = f"annotate-{random_name()}.log"

    with open(os.devnull, "w") as devnull:
        result = subprocess.run(
            ["valgrind", "--tool=callgrind", f"--callgrind-out-file={callgrind_file}"]
            + cmd,
            input=stdin,
            stdout=devnull,
            stderr=devnull,
            text=True,
            timeout=LograderBasicConfig.DEFAULT_EXECUTABLE_TIMEOUT,
        )

    if result.returncode != 0:
        if Path(callgrind_file).is_file():
            os.remove(callgrind_file)
        return []

    result = subprocess.run(
        ["callgrind_annotate", "--auto=yes", callgrind_file],
        stdout=open(annotate_file, "w", encoding="utf-8"),
        stderr=subprocess.DEVNULL,
        text=True,
    )
    os.remove(callgrind_file)

    if result.returncode != 0:
        if Path(annotate_file).is_file():
            os.remove(annotate_file)
        return []

    with open(annotate_file, "r", encoding="utf-8", errors="ignore") as f:
        annotate_output = f.read()
    os.remove(annotate_file)

    return CallgrindOutput(annotate_output).get_calls()


def usr_time(
    cmd: List[Path | str], stdin: Optional[str] = None
) -> ExecutionTimeSummary:
    if sys.platform.startswith("win"):
        return ExecutionTimeSummary()

    time_file = f"time-{random_name()}.log"
    with open(os.devnull, "w") as devnull:
        result = subprocess.run(
            [
                "/usr/bin/time",
                "-f",
                "user_cpu_time=%U\nsystem_cpu_time=%S\ntotal_cpu_time=%e\npercent_cpu_utilization=%P\npeak_physical_memory_usage=%M\nnum_disk_reads=%I\nnum_disk_writes=%O\nnum_major_page_faults=%F\nnum_minor_page_faults=%R\nnum_pages_swapped_to_disk=%W\n",
                "-o",
                time_file,
            ]
            + cmd,
            input=stdin,  # program can still read from stdin
            stdout=devnull,  # hide stdout
            stderr=devnull,
            text=True,
            timeout=LograderBasicConfig.DEFAULT_EXECUTABLE_TIMEOUT,
        )

    if result.returncode != 0:
        if Path(time_file).is_file():
            os.remove(time_file)
        return ExecutionTimeSummary()

    with open(time_file) as f:
        time_stats = f.read()
    os.remove(time_file)

    return TimeOutput(time_stats).get_time()



from datetime import datetime
from importlib import metadata
from pathlib import Path
from typing import List, Literal, Union

from pydantic import BaseModel, Field

# TODO: Add leaderboard support.

ProjectType = Literal["cmake", "cxx-source", "makefile"]


class AssignmentMetadata(BaseModel):
    assignment_name: str
    assignment_authors: List[str]
    assignment_description: str
    assignment_due_date: datetime
    assignment_submit_date: datetime = Field(default_factory=datetime.now)

    @property
    def library_name(self) -> str:
        return "lograder"

    @property
    def library_meta(self) -> metadata.PackageMetadata:
        return metadata.metadata(self.library_name)

    @property
    def library_authors(self) -> List[str]:
        return ["Logan Dapp"]

    @property
    def library_version(self) -> str:
        return metadata.version(self.library_name)


class PreprocessorOutput(BaseModel):
    commands: List[List[Union[str, Path]]]
    stdout: List[str]
    stderr: List[str]

    @property
    def is_successful(self) -> bool:
        return all([cerr == "" for cerr in self.stderr])


class BuilderOutput(BaseModel):
    commands: List[List[str | Path]]
    stdout: List[str]
    stderr: List[str]
    project_type: ProjectType

    @property
    def is_successful(self) -> bool:
        return all([cerr == "" for cerr in self.stderr])



from .._core_exceptions import LograderError


class LograderTestError(LograderError):
    """
    This is the base exception class for all exceptions raised
    by the `lograder.tests` module, for easy error handling.
    """

    pass


class LograderBuildError(LograderTestError):
    """
    This is the base exception class for all exceptions raised
    by the `lograder.tests` module, when the *professor or grader*
    is at fault, for easy error handling.
    """

    pass



from abc import ABC, abstractmethod
from pathlib import Path
from typing import List

from ....tests.registry import TestRegistry
from ....tests.test.interface import ExecutableTestInterface
from ..interface import RunnerInterface, RuntimePrepResults


class ExecutableRunner(RunnerInterface, ABC):
    @abstractmethod
    def get_executable(self) -> List[str | Path]:
        pass

    def prep_tests(self) -> RuntimePrepResults:
        tests: List[ExecutableTestInterface] = []
        for test in TestRegistry.iterate():
            if isinstance(test, ExecutableTestInterface):
                test.set_target(self.get_executable())
                tests.append(test)
            else:
                raise TypeError
        return RuntimePrepResults(tests)



from __future__ import annotations

from typing import List, Optional

from pydantic import BaseModel, Field

from ...output.formatters.default import (
    DefaultBuildOutputFormatter,
    DefaultExecutableTestCaseFormatter,
    DefaultMetadataFormatter,
    DefaultPreprocessorOutputFormatter,
    DefaultRuntimeSummaryFormatter,
)
from ...output.formatters.interfaces import (
    BuildOutputFormatterInterface,
    ExecutableTestFormatterInterface,
    MetadataFormatterInterface,
    PreprocessorOutputFormatterInterface,
    RuntimeSummaryFormatterInterface,
)
from ...output.raw_json.assignment import AssignmentJSON
from ...output.raw_json.test_case import TestCaseJSON
from ...tests.test.interface import ExecutableTestInterface
from .types import AssignmentMetadata, BuilderOutput, PreprocessorOutput


class AssignmentSummary(BaseModel):
    model_config = {"arbitrary_types_allowed": True}

    metadata: AssignmentMetadata
    preprocessor_output: PreprocessorOutput
    build_output: BuilderOutput
    test_cases: List[ExecutableTestInterface]

    metadata_fmt: MetadataFormatterInterface = Field(
        default_factory=DefaultMetadataFormatter, exclude=True
    )
    preprocessor_output_fmt: PreprocessorOutputFormatterInterface = Field(
        default_factory=DefaultPreprocessorOutputFormatter, exclude=True
    )
    build_output_fmt: BuildOutputFormatterInterface = Field(
        default_factory=DefaultBuildOutputFormatter, exclude=True
    )
    runtime_summary_fmt: RuntimeSummaryFormatterInterface = Field(
        default_factory=DefaultRuntimeSummaryFormatter, exclude=True
    )
    test_case_fmt: ExecutableTestFormatterInterface = Field(
        default_factory=DefaultExecutableTestCaseFormatter, exclude=True
    )

    @classmethod
    def set_formatters(
        cls,
        *,
        metadata: Optional[MetadataFormatterInterface] = None,
        preprocessor_output: Optional[PreprocessorOutputFormatterInterface] = None,
        build_output: Optional[BuildOutputFormatterInterface] = None,
        runtime_summary: Optional[RuntimeSummaryFormatterInterface] = None,
        test_case: Optional[ExecutableTestFormatterInterface] = None,
    ):
        if metadata is not None:
            cls.metadata_fmt = metadata
        if preprocessor_output is not None:
            cls.preprocessor_output_fmt = preprocessor_output
        if build_output is not None:
            cls.build_output_fmt = build_output
        if runtime_summary is not None:
            cls.runtime_summary_fmt = runtime_summary
        if test_case is not None:
            cls.test_case_fmt = test_case

    def get_assignment_text(self):
        return (
            f"\n{self.metadata_fmt.format(self.metadata)}"
            f"{self.preprocessor_output_fmt.format(self.preprocessor_output)}\n\n"
            f"{self.build_output_fmt.format(self.build_output)}\n\n"
            f"{self.runtime_summary_fmt.format(self.test_cases)}\n\n"
        )

    def get_score_multiplier(self):
        total_score = sum([test_case.get_weight() for test_case in self.test_cases])
        return 100.0 / total_score if total_score else 0.0

    def get_raw(self) -> AssignmentJSON:
        return AssignmentJSON(
            output=self.get_assignment_text(),
            visibility="visible",
            tests=[
                TestCaseJSON(
                    name=test_case.get_name(),
                    output=self.test_case_fmt.format(test_case),
                    score=self.get_score_multiplier()
                    * test_case.get_weight()
                    * test_case.get_successful()
                    * test_case.get_penalty(),
                    max_score=self.get_score_multiplier() * test_case.get_weight(),
                )
                for test_case in self.test_cases
            ],
            leaderboard=None,  # TODO: Add leaderboard support.
        )



from abc import ABC, abstractmethod
from typing import Union


class RendererInterface(ABC):
    @abstractmethod
    def render(self) -> str:
        pass


class ProcessStep(RendererInterface, ABC):
    pass


class ContextRenderer(RendererInterface):
    _prefix: str = ""
    _suffix: str = ""
    _empty: str = ""

    def __init_subclass__(cls, *, prefix: str, suffix: str, empty: str):
        cls._prefix = prefix
        cls._suffix = suffix
        cls._empty = empty

    def __init__(self, content: Union[RendererInterface, str, None]):
        if content is None:
            content = ""
        self._content: Union[RendererInterface, str] = content

    def get_content(self) -> str:
        return (
            self._content if isinstance(self._content, str) else self._content.render()
        )

    @classmethod
    def get_prefix(cls):
        return cls._prefix

    @classmethod
    def get_suffix(cls):
        return cls._suffix

    @classmethod
    def get_empty(cls):
        return cls._empty

    def render(self) -> str:
        content = self.get_content()
        return (
            f"{self.get_prefix()}{content}{self.get_suffix()}"
            if content
            else self.get_empty()
        )



from colorama import Fore


class LograderMessageConfig:
    DEFAULT_BUILD_ERROR_OVERRIDE_MESSAGE: str = "BUILD WAS UNSUCCESSFUL."
    DEFAULT_BUILD_ERROR_EXECUTABLE_NAME: str = (
        f"{Fore.RED}<NO EXECUTABLE GENERATED>{Fore.RESET}"
    )

    @classmethod
    def set(cls, key: str, value: str):
        if hasattr(cls, key):
            setattr(cls, key, value)



from ...common.types import PreprocessorOutput, ProjectType
from ..interface import (
    BuilderInterface,
    BuilderOutput,
    BuildResults,
    PreprocessorInterface,
    PreprocessorResults,
)


class TrivialPreprocessor(PreprocessorInterface):
    def validate(self):
        return True

    def preprocess(self) -> PreprocessorResults:
        return PreprocessorResults(
            output=PreprocessorOutput(commands=[], stdout=[], stderr=[])
        )


class TrivialBuilder(BuilderInterface):
    def __init__(self, project_type: ProjectType):
        super().__init__()
        self._project_type = project_type

    def build(self) -> BuildResults:
        return BuildResults(
            output=BuilderOutput(
                commands=[],
                stdout=[],
                stderr=[],
                project_type=self._project_type,
            )
        )



import pytest

from lograder import hello_world


@pytest.mark.description("Testing to see if package, `lograder`, built successfully.")
def test_build():
    assert hello_world() == "Hello world! ~from `lograder`!"



from typing import Any, Sequence

from .exceptions import MismatchedSequenceLengthError, NonSingleArgumentSpecifiedError


def validate_common_size(**seqs: Sequence[Any]):
    seq_lens = [len(seq) for _, seq in seqs.items()]
    if not seq_lens:
        return
    initial_seq_len = seq_lens[0]
    if not all([initial_seq_len == seq_len for seq_len in seq_lens]):
        raise MismatchedSequenceLengthError(**seqs)


def validate_unique_argument(**kwargs: Any):
    if sum([val is None for val in kwargs.values()]) != 1:
        raise NonSingleArgumentSpecifiedError(**kwargs)



from __future__ import annotations

import json
from abc import ABC, abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING, List, Optional, Sequence

from ...common.types import FilePath
from ...static import LograderBasicConfig, LograderMessageConfig
from ...tests.common.exceptions import TestNotRunError

if TYPE_CHECKING:
    from ...tests.test.interface import ExecutableTestInterface

from ..common.assignment import AssignmentSummary, BuilderOutput, PreprocessorOutput
from .types import AssignmentMetadata


class PreprocessorResults:
    def __init__(self, output: PreprocessorOutput):
        # The reason why this is a class is in case we want to "expand" later.
        self._output = output

    def get_output(self) -> PreprocessorOutput:
        return self._output


class BuildResults:
    def __init__(self, output: BuilderOutput):
        self._output = output

    def get_output(self) -> BuilderOutput:
        return self._output


class ExecutableBuildResults(BuildResults):
    def __init__(self, executable: FilePath, output: BuilderOutput):
        super().__init__(output)
        self._executable = Path(executable)

    def get_executable(self) -> Path:
        return self._executable


class RuntimePrepResults:  # dummy class to allow distinguishing between pre- and post-run.
    def __init__(
        self,
        results: (
            RuntimePrepResults | RuntimeResults | Sequence[ExecutableTestInterface]
        ),
    ):
        if isinstance(results, RuntimePrepResults):
            self._results = results.get_test_cases()
        else:
            self._results = list(results)

    def get_test_cases(self) -> List[ExecutableTestInterface]:
        return self._results


class RuntimeResults(RuntimePrepResults):
    def get_test_cases(self) -> List[ExecutableTestInterface]:
        for result in self._results:
            if not result.is_executed():
                raise TestNotRunError(result.get_name())
        return self._results


class ProcessInterface(ABC):
    _linked_preprocessors: set[PreprocessorInterface] = set()
    _linked_builders: set[BuilderInterface] = set()
    _linked_runners: set[RunnerInterface] = set()

    @classmethod
    def register_builder(cls, builder: BuilderInterface):
        cls._linked_builders.add(builder)

    @classmethod
    def register_preprocessor(cls, preprocessor: PreprocessorInterface):
        cls._linked_preprocessors.add(preprocessor)

    @classmethod
    def register_runner(cls, runner: RunnerInterface):
        cls._linked_runners.add(runner)

    @classmethod
    def is_build_successful(cls) -> bool:
        for builder in cls._linked_builders:
            if builder.is_build_error():
                return False
        return True

    @classmethod
    def is_preprocessor_successful(cls) -> bool:
        return cls.get_validation_multiplier() == 1.0

    @classmethod
    def get_validation_multiplier(cls) -> float:
        scores = [prep.get_validation_penalty() for prep in cls._linked_preprocessors]
        score = 1.0
        for score in scores:
            score *= score
        return score

    def __eq__(self, other):
        return id(self) == id(
            other
        )  # default "is" comparison, but it's here in case it changes.

    def __hash__(self):
        return hash(id(self))


class PreprocessorInterface(ProcessInterface, ABC):
    def __init__(self):
        super().__init__()
        ProcessInterface.register_preprocessor(self)
        self._validation_penalty: float = 0.0

    def set_validation_penalty(self, penalty: float):
        self._validation_penalty = penalty

    def get_validation_penalty(self) -> float:
        return self._validation_penalty

    def get_individual_validation_multiplier(self) -> float:
        return 1.0 - min(max(self.get_validation_penalty(), 0.0), 1.0)

    def check(self):
        if self.validate():
            return
        self.set_validation_penalty(1.0)

    @abstractmethod
    def validate(self) -> bool:
        pass

    @abstractmethod
    def preprocess(self) -> PreprocessorResults:
        pass


class BuilderInterface(ProcessInterface, ABC):
    def __init__(self):
        super().__init__()
        ProcessInterface.register_builder(self)
        self._build_error: bool = False

    def set_build_error(self, build_error: bool):
        self._build_error = build_error

    def is_build_error(self) -> bool:
        return self._build_error

    @abstractmethod
    def build(self) -> BuildResults:
        pass


class ExecutableBuilderInterface(BuilderInterface, ABC):
    @abstractmethod
    def build(self) -> ExecutableBuildResults:
        pass


class RunnerInterface(ProcessInterface, ABC):
    def __init__(self):
        super().__init__()
        ProcessInterface.register_runner(self)
        self._wrap_args: bool = False
        self._cwd: Optional[Path] = None

    def set_cwd(self, cwd: Path):
        self._cwd = cwd

    def set_wrap_args(self, wrap_args: bool = True):
        self._wrap_args = wrap_args

    def run_tests_auto(self) -> RuntimeResults:
        results: RuntimePrepResults = self.prep_tests()
        for test_case in results.get_test_cases():
            if not self.is_build_successful():
                test_case.force_unsuccessful()
                test_case.override_output(
                    LograderMessageConfig.DEFAULT_BUILD_ERROR_OVERRIDE_MESSAGE,
                    LograderMessageConfig.DEFAULT_BUILD_ERROR_OVERRIDE_MESSAGE,
                )
            else:
                test_case.run(wrap_args=self._wrap_args, working_directory=self._cwd)
        return RuntimeResults(results)

    @abstractmethod
    def prep_tests(self) -> RuntimePrepResults:
        pass


class DispatcherInterface(ABC):
    def run(
        self, out_path: Path = LograderBasicConfig.DEFAULT_RESULT_PATH
    ) -> AssignmentSummary:
        metadata = self.metadata()
        prep = self.preprocess()
        build = self.build()
        runtime_results = self.run_tests()

        summary = AssignmentSummary(
            metadata=metadata,
            preprocessor_output=prep.get_output(),
            build_output=build.get_output(),
            test_cases=runtime_results.get_test_cases(),
        )
        out_path.parent.mkdir(parents=True, exist_ok=True)
        out_path.write_text(json.dumps(summary.model_dump()))
        return summary

    @abstractmethod
    def metadata(self) -> AssignmentMetadata:
        pass

    @abstractmethod
    def preprocess(self) -> PreprocessorResults:
        pass

    @abstractmethod
    def build(self) -> ExecutableBuildResults:
        pass

    @abstractmethod
    def run_tests(self) -> RuntimeResults:
        pass



from .assignment import AssignmentSummary, BuilderOutput, PreprocessorOutput
from .interface import (
    BuilderInterface,
    DispatcherInterface,
    ExecutableBuildResults,
    PreprocessorInterface,
    PreprocessorResults,
    RunnerInterface,
    RuntimePrepResults,
    RuntimeResults,
)
from .templates import CLIBuilder, ExecutableRunner, TrivialBuilder, TrivialPreprocessor

__all__ = [
    "TrivialBuilder",
    "TrivialPreprocessor",
    "AssignmentSummary",
    "CLIBuilder",
    "ExecutableRunner",
    "BuilderInterface",
    "PreprocessorInterface",
    "RunnerInterface",
    "DispatcherInterface",
    "BuilderOutput",
    "ExecutableBuildResults",
    "PreprocessorResults",
    "PreprocessorOutput",
    "RuntimePrepResults",
    "RuntimeResults",
]



import re
import sys
from datetime import datetime
from pathlib import Path
from typing import List, Optional

from ...common.utils import random_name
from ...static import LograderBasicConfig
from ..common import (
    CLIBuilder,
    DispatcherInterface,
    ExecutableBuildResults,
    ExecutableRunner,
    PreprocessorInterface,
    PreprocessorResults,
    RuntimeResults,
    TrivialPreprocessor,
)
from ..common.exceptions import (
    CMakeListsNotFoundError,
)
from ..common.file_operations import bfs_walk, is_cmake_file, is_valid_target
from ..common.types import AssignmentMetadata


class CMakeDispatcher(CLIBuilder, ExecutableRunner, DispatcherInterface):

    # For finding student project target in `cmake --build <...> --target help`
    TARGET_PATTERN = re.compile(r"^\.\.\.\s+([a-zA-Z0-9_\-.]+)", re.MULTILINE)

    def __init__(
        self,
        *,
        assignment_name: str,
        assignment_authors: List[str],
        assignment_description: str,
        assignment_due_date: datetime,
        project_root: Path = LograderBasicConfig.DEFAULT_SUBMISSION_PATH,
        preprocessor: PreprocessorInterface = TrivialPreprocessor(),
    ):
        super().__init__(build_type="cmake")
        self._metadata = AssignmentMetadata(
            assignment_name=assignment_name,
            assignment_authors=assignment_authors,
            assignment_description=assignment_description,
            assignment_due_date=assignment_due_date,
        )

        self._project_root: Path = Path(project_root)
        self._build_directory: Path = self._project_root / f"build-{random_name()}"

        self._cmake_file: Optional[Path] = None
        for file in bfs_walk(self._project_root):
            if is_cmake_file(file):
                self._cmake_file = file
                break
        if self._cmake_file is None:
            raise CMakeListsNotFoundError

        self._working_directory: Path = self._cmake_file.parent
        self._cmake_target: Optional[str] = None
        self._executable_path: Optional[Path] = None

        self._preprocessor = preprocessor
        self._target: Optional[str] = None

    def get_project_root(self) -> Path:
        return self._project_root

    def get_build_directory(self) -> Path:
        return self._build_directory

    def get_working_directory(self) -> Path:
        return self._working_directory

    def get_executable(self) -> List[str | Path]:
        assert self._target is not None  # mypy
        try:
            return [self.find_executable(self._target)]
        except FileNotFoundError:
            return []

    def find_executable(self, target: str) -> Path:
        build_dir = self.get_build_directory()
        candidates = [
            build_dir / "Debug" / f"{target}.exe",
            build_dir / "Release" / f"{target}.exe",
            build_dir / "Debug" / target,
            build_dir / "Release" / target,
            build_dir / f"{target}.exe",
            build_dir / target,
        ]

        for path in candidates:
            if path.is_file():
                return path

        for path in build_dir.rglob("*"):
            if path.is_file():
                if path.name == target or path.name == f"{target}.exe":
                    return path

        raise FileNotFoundError

    def build(self) -> ExecutableBuildResults:
        cmd: List[str | Path] = [
            "cmake",
            "-S",
            self.get_working_directory(),
            "-B",
            self.get_build_directory(),
        ]
        if sys.platform.startswith("win"):
            cmd += ["-G", "MinGW Makefiles"]

        self.run_cmd(cmd)
        if self.is_build_error():
            return self.get_build_error_output()

        cmd = ["cmake", "--build", self.get_build_directory(), "--target", "help"]
        output = self.run_cmd(cmd)
        if self.is_build_error():
            return self.get_build_error_output()

        targets = self.TARGET_PATTERN.findall(output.stdout[-1])
        if "main" in targets:
            self._target = "main"
        elif "build" in targets:
            self._target = "build"
        elif "demo" in targets:
            self._target = "demo"
        else:
            valid_targets = [target for target in targets if is_valid_target(target)]
            if not valid_targets:
                self.set_build_error(True)
                return self.get_build_error_output()
            self._target = valid_targets[0]

        cmd = [
            "cmake",
            *LograderBasicConfig.DEFAULT_CMAKE_COMPILATION_FLAGS,
            "--build",
            self.get_build_directory(),
            "--target",
            self._target,
            "--",
            "-s",
            "--no-print-directory",
        ]
        output = self.run_cmd(cmd)
        if self.is_build_error():
            return self.get_build_error_output()

        try:
            self._executable_path = self.find_executable(self._target)
        except FileNotFoundError:
            self.set_build_error(True)
            return self.get_build_error_output()

        assert self._executable_path is not None  # mypy
        return ExecutableBuildResults(
            executable=self._executable_path.resolve(), output=output
        )

    def metadata(self) -> AssignmentMetadata:
        return self._metadata

    def preprocess(self) -> PreprocessorResults:
        return self._preprocessor.preprocess()

    def run_tests(self) -> RuntimeResults:
        return self.run_tests_auto()



from pathlib import Path
from typing import List


class LograderBasicConfig:
    DEFAULT_SUBMISSION_PATH: Path = Path("/autograder/submission")
    DEFAULT_RESULT_PATH: Path = Path("/autograder/results/results.json")
    DEFAULT_CXX_STANDARD: str = "c++20"
    DEFAULT_CXX_COMPILATION_FLAGS: List[str] = ["-Wall", "-Wextra", "-Werror"]
    DEFAULT_CMAKE_COMPILATION_FLAGS: List[str] = []
    DEFAULT_EXECUTABLE_TIMEOUT: float = 300.0  # In seconds.

    DEFAULT_TOPIC_BREAK: str = "\n\n"

    @classmethod
    def set(cls, key: str, value: str):
        if hasattr(cls, key):
            setattr(cls, key, value)



from ._core_exceptions import LograderStudentBuildError
from .common.exceptions import (
    CMakeExecutableNotFoundError,
    CMakeListsNotFoundError,
    CMakeTargetNotFoundError,
    CxxSourceBuildError,
    MakefileNotFoundError,
    MakefileRunNotFoundError,
)

__all__ = [
    "LograderStudentBuildError",
    "CMakeListsNotFoundError",
    "CMakeExecutableNotFoundError",
    "CMakeTargetNotFoundError",
    "CxxSourceBuildError",
    "MakefileRunNotFoundError",
    "MakefileNotFoundError",
]



from .comparison_test import ExecutableOutputComparisonTest
from .interface import ExecutableTestInterface, TestInterface

__all__ = [
    "ExecutableOutputComparisonTest",
    "TestInterface",
    "ExecutableTestInterface",
]



from typing import Literal

TextFormat = Literal["text", "html", "simple_format", "md", "ansi"]
Visibility = Literal["visible", "after_due_date", "after_published", "hidden"]
Status = Literal["passed", "failed"]
AscendingOrder = Literal["asc"]
